% sv.P 
%-------------------------------------

:- set_flag(occur_check, on).

/*
reach/5: arguments:
B = semibundle
N = invariant
Kl = intruder legal knowledge w/possible instantiations
Ki = intruder illegal knowledge
F = Varconstraints (initially empty list)
Trace = output trace
*/

reach(B, N, Kl, Ki, L1, Trace) :-
  not(evalformula(N,Kl,Ki,L1,Trace)),
  reverse(Trace,Tr),
  nl,write('trace: '), nl, write_ll(Tr).

% always process 'send' messages first
reach(B,N,Kl,Ki,L,Trace) :-
  member([Strand,[[send(M),_]|_S]],B),
  !,

  diff(B,send(M),B1),
  (
  (Strand = [agent(Ss), apparent_party(A)], \+(\+(A=e)), 
   Strand2 = [agent(e), apparent_party(Ss)],

   reach(B1, N, [M|Kl], Ki, L, [
                                 [Strand2,recv(M)],
                                  [Strand,send(M)]
                                |Trace]));

  (Strand = [agent(Ss), apparent_party(A)], A\=e, 
   Strand2 = [agent(e(A)), apparent_party(Ss)],
   reach(B1, N, Kl, [M|Ki], L, [
                                
                                 [Strand2,recv_interception(M)],
                                 [Strand,send(M)]
                                |Trace]))
  )
.

reach(B,N,Kl,Ki, L1, Trace) :- 
  append(Kl,Ki,T),
  create_const(B, T, CC),	% look for first nodes each strand of bundle
  (

  (
  member([Strand,[[recv(Constr),F]|TT]], CC),	% create constraint
  solve([[Constr|TT]], L),	% try to solve it


  append(L1,L,L2),

  simplify(L2,L3),		% simplify list of constraints
%  not(evalformula(F, Kl, Ki, L3, Trace)),
%  simplify(L3,L4),		% simplify list of constraints

  diff(B, recv(Constr), B1),	

  Strand = [agent(Dest), apparent_party(Apparent)],
  reach(B1, N,Kl,Ki, L3, [[Strand,recv(Constr)]|[[
						   [agent(e(Apparent)),dest(Dest)],
                                                   send(Constr)
                                                  ]|Trace
						 ]
                          ])
  );

  (
  member([Strand,[check(Constr)|TT]], CC),	% create constraint
  solve([[Constr|TT]], L),	% try to solve it
  append(L1,L,L2),
  simplify(L2,L3),		% simplify list of constraints
  diff(B, check(Constr), B1),	% if solved, remove it from strand (possibly repeated!!)
				% notice that B and T are unified against Constr
  reach(B1, N, Kl,Ki, L3, [[Strand,check(Constr)]|Trace])
  )

  ).

evalformula([], _, _, _, _).
evalformula([[]], _, _, _, _).

evalformula(secrecy(F), Kl, Ki, L, Trace) :-
 append(Kl, Ki, T),

 \+(solve([[F|[T]]], L1)),

 nl,write('failed goal: '),
 write(secrecy(F)),nl.

evalformula(not(secrecy(F)), Kl, Ki, L, Trace) :-
 append(Kl, Ki, T),
 solve([[F|[T]]], L),

 nl,write('failed goal: '),
 write(secrecy(F)),nl.

evalformula(generable(F), Kl, Ki, L, Trace) :-
 solve([[F|[Kl]]], L2),

% !,
 nl,write('failed goal: '),
 write(generable(F)),nl.

evalformula(not(generable(F)), Kl, Ki, L, Trace) :-
 not(solve([[F|[Kl]]], L2)),

% !,
 nl,write('failed goal: '),
 write(secrecy(F)),nl.

evalformula(equal(T1,T2), Kl, Ki, L, Trace) :-
 not(not(T1=T2)),
% !,
 nl,write('failed goal: '),
 write(equal(T1,T2)),nl.

evalformula(not(equal(T1,T2)), Kl, Ki, L, Trace) :-
 not(not(T1\=T2)),
% !,
 nl,write('failed goal: '),
 write(notequal(T1,T2)),nl.

evalformula(in(T), Kl, Ki, L, Trace) :-
 member(T, Trace),
% !,
 nl,write('failed goal: '),
 write(in(T)),nl.

evalformula(not(in(T)), Kl, Ki, L, Trace) :-
 nonmember(T, Trace),
% !,
 nl,write('failed goal: '),
 write(not(in(T))),nl.

evalformula(not(and(F1,F2)), Kl, Ki, L, Trace) :-
	evalformula(or(not(F1),not(F2)), Kl, Ki, L, Trace).

evalformula(not(or(F1,F2)), Kl, Ki, L, Trace) :-
	evalformula(and(not(F1),not(F2)), Kl, Ki, L, Trace).

evalformula(imp(F1,F2), Kl, Ki, L, Trace) :-
	evalformula(or(not(F1),F2), Kl, Ki, L, Trace).

evalformula(or(F1,F2), Kl, Ki, L, Trace) :-
	evalformula(F1, Kl, Ki, L, Trace);
	evalformula(F2, Kl, Ki, L, Trace).

evalformula(and(F1,F2), Kl, Ki, L, Trace) :-
	evalformula(F1, Kl, Ki, L, Trace),
	evalformula(F2, Kl, Ki, L, Trace).


/*
simplify/2 arguments: +L1,?L2
L1 = list of constraints (input)
L2 = list of solved constraints (output)

fail if some constraint in L1 is not solvable
*/

simplify(L,L) :-
	all_variable(L).	% all simple constraints, finish

simplify(L,L1) :-
	delete([X|K],L,L2),	
	nonvar(X),
	solve([[X|K]],L3),	% nonsimple constraint, solve it
	append(L2,L3,L4),	
	simplify(L4,L1).	% maybe new non-simple constraint were added


/* all_variable/1 checks if all constraints are simple */
all_variable([]).	
all_variable([[X|_]|Rest]):-var(X), all_variable(Rest).

% diff, remove N from a strand S in the bundle B
diff([[St,[[N,_]|S]]|B],N,[[St,S]|B]).
diff([[St,S]|B],N,[[St,S]|B1]) :-diff(B,N,B1).

% create constraints by choosing only the first nodes of the strands of the form recv(M)

create_const([],_,[]).
create_const([[_,[]] |B],T,C) :- create_const(B,T,C).
create_const([[St,[M|_]] |B],T,[ [St,[M|[T]] ]|C]) :- create_const(B, T, C).

%-------------------------------------

% solve([[expr, termlist],...],Varconstraints)
% apply reduct to each nonsimple constraint, in
% reverse (i.e., chronological) order.
% Build up (possibly empty) list of simple [var,termlist] constraints.
% Note that reduct may cause a var to be instantiated
% on the left side of a prior constraint, so
% tail recursion applies solve again from the beginning.

solve(C,C) :- allvarc(C), !.

solve([[A,T]|C],W1) :- 
   solve(C,V),
   remv(T,T1),
   reduct(A,T1,U),
   append(U,V,W),
   solve(W,W1). 

% reduct(M,T,C) performs one reduction step on
% an active constraint M:T and 
% records replacement constraints in C


reduct(M,T,[[M,T]]) :-
  var(M),!.		% pass over simple constraint

reduct(U=V,_,[]) :-		% (check)
  unify(U,V).

reduct(M,T,[]) :-  	% (un) with constant
  atomic(M),		% always safe
  member(M,T),!.


reduct([A,B],T,[[B,T],[A,T]]) :- !. 	% (pair), always safe

reduct(M/pk(e),T,[[M,T]]) :- !. 	% (sig), always safe

reduct(M,T,[]) :-		% (un)
  member(A,T),
  hunify(M,A).

reduct(sha(M),T,[[M,T]]).  	% (hash)

reduct(pk(A),T,[[A,T]]).	% public-key lookup

reduct(M*K,T,[[M,T],[K,T]]). 	% (penc)

reduct(M+K,T,[[M,T],[K,T]]).	% (senc)

reduct(M,T,[[M,T]]) :-
  do_ksub(T).

reduct(M,T,[[M,T2],[K,T1]]) :-
  do_ksyn(T,T2,K,T1).

% remv removes variables from a term list, if any.
% It also does (split) and (pdec), they're always safe.

remv([],[]) :- !.

remv([A|T],W) :-
  var(A),!,
  remv(T,W).

remv([[A,B]|T],W) :- !,	% (split)
  remv([A,B|T],W).

remv([U*K|T],W) :-
  K==pk(e),!,		% (pdec)
  remv([U|T],W).

remv([A|T],[A|W]) :-
  remv(T,W).

% do_ksub looks for U*V in a term list
% and binds V to pk(e) if possible (and V not already pk(e))
% It fails if there is no instance to bind.

do_ksub([_U*V|_T]) :-
  \+V==pk(e),
  V=pk(e).

do_ksub([_A|T]) :- do_ksub(T).

% do_ksyn looks for U+K in a term list T
% and decrypts it to U.  We also insert the
% new constraint for K with U-K in T1.
% Fails iff there is no symmetric encryption.

do_ksyn([U+K|T],[U,K|T],K,[U-K|T]).

do_ksyn([A|T],[A|T2],K,[A|T1]) :- 
  do_ksyn(T,T2,K,T1).

% allnull tests for empty bundle

allnull([]).

allnull([[]|B]) :- allnull(B).

% allvarc tests for simple constraint set
%    in which all left sides are variables

allvarc([]).

allvarc([[X,_T]|C]) :-
  var(X),
  allvarc(C).

% hunify(M,A) turns A from - to + first if necessary.

hunify(M,U-V) :- unify(M,U+V),!.
hunify(M,A) :- unify(M,A).

%----------------------------------------------------------------------------
% "safe" unification with occurs check from C. Meadows
%----------------------------------------------------------------------------

unify(X,Y) :-
	var(X),var(Y),!,
	X=Y.
unify(X,Y) :-
	atomic(X),!,X=Y.	% atomic includes numbers
unify(X,Y) :-
	atomic(Y),!,X=Y.
unify(X,Y) :-
	var(X),!,
	notOccurs(X,Y),
	X=Y.
unify(X,Y) :-
	var(Y),!,
	notOccurs(Y,X),
	X=Y.
unify(X,Y) :-
	X =.. [A|B],
	Y =.. [A|C],
	list_unify(B,C),!.

list_unify([],[]).
list_unify([A|B],[C|D]) :-
	unify(A,C),
	list_unify(B,D).

notOccurs(X,Y) :- var(Y),!, \+ X == Y.
notOccurs(_X,Y) :- atomic(Y),!.
notOccurs(X,[Y|Z]) :- !,notOccurs(X,Y),notOccurs(X,Z).
notOccurs(X,Y) :- Y =.. [_F|N],notOccurs(X,N).


% pretty print of lists
write_ll([]).
write_ll([H|T]) :- writeq(H),nl,write_ll(T).

